
/*
 * Copyright (C) Igor Sysoev
 * Copyright (C) Nginx, Inc.
 */

#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_http.h>
#include <linux/types.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define NGX_TTM_BASE_CTL (64 + 1024 + 64 + 64 + 64 + 64) /* base */
#define NGX_TTM_SO_GET_PEER (NGX_TTM_BASE_CTL + 1)

typedef struct ttm_peer_s ttm_peer_t;

struct ttm_peer_s
{
    __u16 af;
    __be16 port;
    union {
        __u32 all[4];
        __be32 ip;
        __be32 ip6[4];
        struct in_addr in;
        struct in6_addr in6;
    };
};

typedef struct
{
    ngx_str_t header;
    ngx_flag_t set_real_ipv6;
} ngx_http_realipv6_loc_conf_t;

typedef struct
{
    ngx_str_t remote_ipv6_addr;
} ngx_http_realipv6_ctx_t;

static ngx_int_t ngx_http_realipv6_handler(ngx_http_request_t *r);
static void ngx_http_realipv6_cleanup(void *data);
static void *ngx_http_realipv6_create_loc_conf(ngx_conf_t *cf);
static ngx_int_t ngx_http_realipv6_add_variables(ngx_conf_t *cf);
static ngx_int_t ngx_http_realipv6_init(ngx_conf_t *cf);
static ngx_int_t ngx_http_realipv6_remote_addr_variable(ngx_http_request_t *r,
                                                        ngx_http_variable_value_t *v, uintptr_t data);

static ngx_command_t ngx_http_realipv6_commands[] = {

    {ngx_string("set_real_ipv6_header"),
     NGX_HTTP_MAIN_CONF | NGX_HTTP_SRV_CONF | NGX_HTTP_LOC_CONF | NGX_CONF_TAKE1,
     ngx_conf_set_str_slot,
     NGX_HTTP_LOC_CONF_OFFSET,
     offsetof(ngx_http_realipv6_loc_conf_t, header),
     NULL},

    {ngx_string("set_real_ipv6"),
     NGX_HTTP_MAIN_CONF | NGX_HTTP_SRV_CONF | NGX_HTTP_LOC_CONF | NGX_CONF_FLAG,
     ngx_conf_set_flag_slot,
     NGX_HTTP_LOC_CONF_OFFSET,
     offsetof(ngx_http_realipv6_loc_conf_t, set_real_ipv6),
     NULL},

    ngx_null_command};

static ngx_http_module_t ngx_http_realipv6_module_ctx = {
    ngx_http_realipv6_add_variables, /* preconfiguration */
    ngx_http_realipv6_init,          /* postconfiguration */

    // ngx_http_realipv6_add_variables, /* create main configuration */
    NULL,
    NULL, /* init main configuration */

    NULL, /* create server configuration */
    NULL, /* merge server configuration */

    ngx_http_realipv6_create_loc_conf, /* create location configuration */
    NULL                               /* merge location configuration */
};

ngx_module_t ngx_http_realipv6_module = {
    NGX_MODULE_V1,
    &ngx_http_realipv6_module_ctx, /* module context */
    ngx_http_realipv6_commands,    /* module directives */
    NGX_HTTP_MODULE,               /* module type */
    NULL,                          /* init master */
    NULL,                          /* init module */
    NULL,                          /* init process */
    NULL,                          /* init thread */
    NULL,                          /* exit thread */
    NULL,                          /* exit process */
    NULL,                          /* exit master */
    NGX_MODULE_V1_PADDING};

static ngx_http_variable_t ngx_http_realipv6_vars[] = {

    {ngx_string("x_real_ipv6"), NULL,
     ngx_http_realipv6_remote_addr_variable, 0, 0, 0},

    {ngx_null_string, NULL, NULL, 0, 0, 0}};

static ngx_int_t
ngx_http_realipv6_handler(ngx_http_request_t *r)
{
    u_char *p;
    socklen_t len;
    size_t plen;
    ngx_connection_t *c;
    ngx_http_realipv6_loc_conf_t *rlcf;
    ngx_pool_cleanup_t *cln;
    ngx_http_realipv6_ctx_t *ctx;

    rlcf = ngx_http_get_module_loc_conf(r, ngx_http_realipv6_module);

    if (!rlcf->set_real_ipv6)
    {
        return NGX_DECLINED;
    }
    c = r->connection;
    ttm_peer_t peer;
    len = sizeof(peer);
    if (getsockopt(c->fd, IPPROTO_IP, NGX_TTM_SO_GET_PEER, &peer, &len) < 0)
    {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                      "nginx getsockopt failed");
        return NGX_DECLINED;
    }
    char addr[NGX_SOCKADDR_STRLEN];
    const char *ret = inet_ntop(peer.af, &peer.all, addr, sizeof(addr));
    if (ret == NULL)
    {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                      "nginx inet_ntop failed");
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }

    cln = ngx_pool_cleanup_add(r->pool, sizeof(ngx_http_realipv6_ctx_t));
    if (cln == NULL)
    {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    ctx = cln->data;
    ngx_http_set_ctx(r, ctx, ngx_http_realipv6_module);

    c = r->connection;
    plen = ngx_strlen(ret);
    p = ngx_pnalloc(c->pool, plen);
    if (p == NULL)
    {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
    ngx_memcpy(p, ret, plen);
    cln->handler = ngx_http_realipv6_cleanup;
    ctx->remote_ipv6_addr.len = plen;
    ctx->remote_ipv6_addr.data = p;
    return NGX_DECLINED;
}

static void
ngx_http_realipv6_cleanup(void *data)
{
    return;
}

static void *
ngx_http_realipv6_create_loc_conf(ngx_conf_t *cf)
{
    ngx_http_realipv6_loc_conf_t *conf;

    conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_realipv6_loc_conf_t));
    if (conf == NULL)
    {
        return NULL;
    }

    ngx_str_null(&conf->header);
    conf->set_real_ipv6 = NGX_CONF_UNSET;

    return conf;
}

static ngx_int_t
ngx_http_realipv6_add_variables(ngx_conf_t *cf)
{
    ngx_http_variable_t *var, *v;

    for (v = ngx_http_realipv6_vars; v->name.len; v++)
    {
        var = ngx_http_add_variable(cf, &v->name, v->flags);
        if (var == NULL)
        {
            return NGX_ERROR;
        }

        var->get_handler = v->get_handler;
        var->data = v->data;
    }

    return NGX_OK;
}

static ngx_int_t
ngx_http_realipv6_init(ngx_conf_t *cf)
{
    ngx_http_handler_pt *h;
    ngx_http_core_main_conf_t *cmcf;

    cmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_core_module);

    h = ngx_array_push(&cmcf->phases[NGX_HTTP_POST_READ_PHASE].handlers);
    if (h == NULL)
    {
        return NGX_ERROR;
    }

    *h = ngx_http_realipv6_handler;

    return NGX_OK;
}

static ngx_int_t
ngx_http_realipv6_remote_addr_variable(ngx_http_request_t *r,
                                       ngx_http_variable_value_t *v, uintptr_t data)
{
    ngx_str_t *remote_ipv6_addr;
    ngx_pool_cleanup_t *cln;
    ngx_http_realipv6_ctx_t *ctx;

    ctx = ngx_http_get_module_ctx(r, ngx_http_realipv6_module);

    if (ctx == NULL && (r->internal || r->filter_finalize))
    {
        /*
         * if module context was reset, the original address
         * can still be found in the cleanup handler
         */
        for (cln = r->pool->cleanup; cln; cln = cln->next)
        {
            if (cln->handler == ngx_http_realipv6_cleanup)
            {
                ctx = cln->data;
                break;
            }
        }
    }

    // remote_ipv6_addr = ctx->remote_ipv6_addr;
    remote_ipv6_addr = ctx ? &ctx->remote_ipv6_addr : &r->connection->addr_text;
    v->len = remote_ipv6_addr->len;
    v->valid = 1;
    v->no_cacheable = 0;
    v->not_found = 0;
    v->data = remote_ipv6_addr->data;

    return NGX_OK;
}
